<?php
namespace Models;

class Chz{
	public $db, $f3, $sets, $web, $chz_url, $cdn_table;
	public $cash_time=6; // в часах

	public $errors=array(
		1=>'ошибка валидации КМ',
		2=>'КМ не содержит GTIN',
		3=>'КМ не содержит серийный номер',
		4=>'КМ содержит недопустимые символы',
		5=>'Ошибка верификации крипто-подписи КМ (формат крипто-подписи не соответствует типу КМ)',
		6=>'Ошибка верификации крипто-подписи КМ (криптоподпись не валидная)',
		7=>'Ошибка верификации крипто-подписи КМ (крипто-ключ не валиден)',
		8=>'КМ не прошел верификацию в стране эмитента',
		9=>'Найденные AI в КМ не поддерживаются',
		10=>'КМ не найден в ГИС МТ',
		11=>'КМ не найден в трансгране'
	);
	
	public function __construct($db, $f3, $sets) {
		$this->cdn_table=new \DB\SQL\Mapper($db, 'chz_cdn');
		$this->db=$db;
		$this->f3=$f3;
		$this->sets=$sets;
		$this->web=\Web::instance();
		$this->chz_url='https://cdn.crpt.ru'; // Боевой
		if($this->f3->get('HOST')=='kassa.loc') {
			$this->chz_url='https://markirovka.sandbox.crptech.ru';
		}
	}

	// Запрос CDN площадок
	function get_cdn() {
		if(!isset($this->sets['token_chz']) || !$this->sets['token_chz']) {
			return array(
				'success'=>false,
				'code'=>401,
				'check_result'=>-1,
				'txt'=>'checktoken_false_autorization'
			);
		}
		$options=array(
			'method'=>'GET',
			'header'=>array(
				'Content-Type: application/json',
				'X-API-KEY: '.$this->sets['token_chz']
			),
			'timeout'=>1.5
		);
		$url=$this->chz_url.'/api/v4/true-api/cdn/info';
		$res=$this->web->request($url, $options);
		$json=json_decode($res['body'], true);
		if(mb_strpos($res['headers'][0], '401')>0) {
			return array(
				'success'=>false,
				'code'=>'401',
				'txt'=>$json['error_message']??'checktoken_false_autorization'
			);
		}
		if(mb_strpos($res['headers'][0], '200')===false) {
			return array(
				'success'=>false,
				'code'=>$res['headers'][0]
			);
		}

		if(!is_array($json)) {
			return array(
				'success'=>false,
				'code'=>$json['code'],
				'txt'=>$json['error_message']
			);
		}

		if((int)$json['code']!=0) {
			return array(
				'success'=>false
			);
		}

		return array(
			'success'=>true,
			'hosts'=>array_column($json['hosts'], 'host')
		);
	}

	// Проверка CDN площадки
	function host_check($host) {
		$time=microtime(true);
		$options=array(
			'method'=>'GET',
			'header'=>array(
				'Content-Type: application/json',
				'X-API-KEY: '.$this->sets['token_chz']
			),
			'timeout'=>1.5
		);
		$url=$host.'/api/v4/true-api/cdn/health/check';
		$res=$this->web->request($url, $options);
		$json=json_decode($res['body'], true);

		if(mb_strpos($res['headers'][0], '401')>0) {
			return array(
				'success'=>false,
				'code'=>'401',
				'txt'=>$json['error_message']??'checktoken_false_autorization'
			);
		}

		$time=(microtime(true)-$time);
		if(mb_strpos($res['headers'][0], '200')===false) {
			return array(
				'code'=>-1,
				'host'=>$host,
				'time_request'=>$time
			);
		}
		if(!is_array($json)) {
			return array(
				'code'=>-1,
				'host'=>$host,
				'time_request'=>$time
			);
		}

		$json['host']=$host;
		$json['time_request']=$time;
		return $json;
	}

	function load_cdn($query=null, $options=null) {
		$cdn=array();
		$this->cdn_table->reset();
		$this->cdn_table->load($query, $options);
		if($this->cdn_table->dry()) return $cdn;
		do {
			$cdn[$this->cdn_table->host]=$this->cdn_table->cast();
		}
		while($this->cdn_table->skip());
		return $cdn;
	}

	function save_cdn($cdn) {
		if(!$cdn['host']) return;
		$this->cdn_table->reset();
		$this->cdn_table->load(array('host=?', $cdn['host']), array('limit'=>1));
		$this->cdn_table->copyFrom($cdn);
		$this->cdn_table->save();
	}

	// Актуализация списка площадок
	function cdn_actually($cdn) {
		$cdn_actually=$this->get_cdn();
		if($cdn_actually['success']) {
			foreach($cdn_actually['hosts'] as $host) {
				if(!isset($cdn[$host])) {
					$cdn[$host]=array(
						'host'=>$host,
						'code'=>0,
						'avgTimeMs'=>0,
						'time_request'=>0,
						'time_update'=>0
					);
					$this->save_cdn($cdn[$host]);
				}
			}
		}
		return $cdn_actually;
	}

	// проверка CDN
	function check_cdn() {
		$now=time();
		// Список загруженных площадок
		$cdn=$this->load_cdn();

		// Сколько часов назад была актуализация списка площадок
		$d=($now-$this->sets['chz_cdn_update'])/3600;

		// Если прошло больше 6 часов или нет списка площадок, то нужно запросить список площадок
		if($d>=$this->cash_time || empty($cdn)) {
			$res=$this->cdn_actually($cdn);

			if($res['success']) {
				$sql="INSERT OR REPLACE INTO `settings` (`name`, `value`) VALUES ('chz_cdn_update', ?)";
				$val=array($now);
				$this->db->exec($sql, $val);
			}
			else {
				return $res;
			}
			$cdn=$this->load_cdn();
		}
		
		if(empty($cdn)) {
			return array(
				'success'=>false
			);
		}

		// Смотрим какие площадки нужно перепроверить
		foreach ($cdn as $c) {
			if($c['time_update']<$now) {
				$res=$this->host_check($c['host']);
				if($res['code']<0) continue;
				if($res['code']==429 || $res['code']>=500) {
					$res['time_update']=$now+900;
				}
				else {
					$res['time_update']=$now+3600*$this->cash_time;
				}
				$this->save_cdn($res);
			}
		}

		return array(
			'success'=>true
		);
	}

	// Запрос проверки марки
	function check_mark_code($host, $mark) {
		$mark=str_replace($this->sets['gs1_symbols'], '\u001d', $mark);
		$mark=str_replace('', '\u001d', $mark);
		$vars=array(
			'codes'=>array($mark)
		);
		$options=array(
			'method'=>'POST',
			'header'=>array(
				'Content-Type: application/json',
				'X-API-KEY: '.$this->sets['token_chz']
			),
			'content'=>json_encode($vars),
			'timeout'=>1.5
		);

		$url=$host.'/api/v4/true-api/codes/check';
		$res=$this->web->request($url, $options);
		$json=json_decode($res['body'], true);

		if(mb_strpos($res['headers'][0], '401')>0) {
			return array(
				'success'=>false,
				'code'=>'401',
				'txt'=>$json['error_message']??'checktoken_false_autorization'
			);
		}

		if(mb_strpos($res['headers'][0], '200')===false) {
			return array(
				'code'=>-1,
			);
		}

		if(!is_array($json)) {
			return array(
				'code'=>-1,
			);
		}

		return $json;
	}

	// Проверка марки
	function check_mark($line, $mark) {
		if(!isset($this->sets['token_chz']) || !$this->sets['token_chz']) {
			return array(
				'success'=>false,
				'code'=>401,
				'check_result'=>-1,
				'txt'=>'checktoken_false_autorization'
			);
		}
		// Проверяем площадки
		$res=$this->check_cdn();
		if((int)$res['code']==401) {
			$res['check_result']=-1;
			return $res;
		}
		// Получаем список доступных площадок, начиная от самой близкой
		$cdn=$this->load_cdn(array('code=0'), array('order'=>'time_request ASC'));
		if(empty($cdn)) {
			return array(
				'code'=>-1,
			);
		}

		// проверяем марку. Если площадка недоступна, то пробуем на других
		$now=time();
		$res=array('code'=>-1);
		foreach($cdn as $c) {
			$res=$this->check_mark_code($c['host'], $mark);
			if((int)$res['code']==401) {
				$res['check_result']=-1;
				return $res;
			}
			if($res['code']==0) break;
			$c['code']=$res['code'];
			$c['time_update']=$time+900;
			$this->save_cdn($c);
		}

		if($res['code']<0) {
			return array(
				'success'=>true,
				'check_result'=>0,
			);
		}

		$success=true;

		if($res['code']>0) {
			$success=false;
			$txt=$res['description']??'error';
		}
		$mark_res=$res['codes'][0];

		if($mark_res['errorCode']>0) {
			$success=false;
			$txt=$this->errors[$mark_res['errorCode']]??$mark_res['message'];
		}
		else {
			if($mark_res['found']===false) {
				$success=false;
				$txt.='Код идентификации не найден в ГИС МТ<br>';
			}

			if($mark_res['utilised']===false) {
				$success=false;
				$txt.='Код маркировки эмитирован, но нет информации о его нанесении<br>';
			}

			if($mark_res['verified']===false) {
				$success=false;
				$txt.='Не пройдена криптографическая проверка кода маркировки<br>';
			}

			if($mark_res['sold']===true) {
				$success=false;
				$txt.='Код идентификации выведен из оборота<br>';
			}

			if($mark_res['isBlocked']===true) {
				$success=false;
				$txt.='Код идентификации заблокирован по решению ОГВ<br>';
			}

			if($mark_res['realizable']===false) {
				if($res['grayZone']!==true || ($line['marking_group']!=3 && $line['marking_group']!=12)) {
					$success=false;
					$txt.='Нет информации о вводе в оборот кода идентификации<br>';
				}
			}

			if(isset($mark_res['expireDate']) && strtotime($mark_res['expireDate'])<=$now) {
				$success=false;
				$txt.='Товар с истекшим сроком годности<br>';
			}
		}

		if(isset($mark_res['smp']) || isset($mark_res['mrp'])) {
			$price=$line['price']+($line['price']*($line['vat']/100))*max(0, $line['vat_type']);
			if(isset($mark_res['smp']) && $price*100<$mark_res['smp']) {
				//$success=false;
				//$txt='Продажа ниже минимальной цены. Минимальная цена '.number_format($mark_res['smp']/100, 2, '.', '');
				if($line['vat_type']>0) {
					$mark_res['set_price']=($mark_res['smp']/(100+$line['vat'])*100)/100;
				}
				else {
					$mark_res['set_price']=$mark_res['smp']/100;
				}
			}

			if($success==true && isset($mark_res['mrp']) && $price*100>$mark_res['mrp']) {
				//$success=false;
				//$txt='Продажа выше максимальной цены. Максимальная цена '.number_format($mark_res['mrp']/100, 2, '.', '');
				if($line['vat_type']>0) {
					$mark_res['set_price']=($mark_res['mrp']/(100+$line['vat'])*100)/100;
				}
				else {
					$mark_res['set_price']=$mark_res['mrp']/100;
				}
			}
		}
		if($success==true && !isset($mark_res['mrp']) && ($line['marking_group']==3 || $line['marking_group']==12)) {
			$mrc=mb_substr($mark, 21, 4);
			$mrc=$this->mrc($mrc);
			if($mrc<$price) {
				//$success=false;
				//$txt='Продажа выше максимальной цены. Максимальная цена '.number_format($mrc/100, 2, '.', '');
				if($line['vat_type']>0) {
					$mark_res['set_price']=$mrc/(100+$line['vat'])*100;
				}
				else {
					$mark_res['set_price']=$mrc;
				}
			}
		}

		if($success===false) {
			return array(
				'success'=>false,
				'check_result'=>-1,
				'error_type'=>'chz',
				'txt'=>$txt,
				'result'=>$mark_res,
				'reqId'=>$res['reqId'],
				'reqTimestamp'=>$res['reqTimestamp']
			);
		}

		return array(
			'success'=>true,
			'check_result'=>1,
			'result'=>$mark_res,
			'reqId'=>$res['reqId'],
			'reqTimestamp'=>$res['reqTimestamp']
		);
	}

	// Находим МРЦ - максимальная розничная цена
	function mrc($mrc) {
		$result=0;
		$s=array(
			'A'=>0,'B'=>1,'C'=>2,'D'=>3,'E'=>4,'F'=>5,'G'=>6,'H'=>7,'I'=>8,'J'=>9,'K'=>10,
			'L'=>11,'M'=>12,'N'=>13,'O'=>14,'P'=>15,'Q'=>16,'R'=>17,'S'=>18,'T'=>19,'U'=>20,
			'V'=>21,'W'=>22,'X'=>23,'Y'=>24,'Z'=>25,'a'=>26,'b'=>27,'c'=>28,'d'=>29,'e'=>30,
			'f'=>31,'g'=>32,'h'=>33,'i'=>34,'j'=>35,'k'=>36,'l'=>37,'m'=>38,'n'=>39,'o'=>40,
			'p'=>41,'q'=>42,'r'=>43,'s'=>44,'t'=>45,'u'=>46,'v'=>47,'w'=>48,'x'=>49,'y'=>50,
			'z'=>51,'0'=>52,'1'=>53,'2'=>54,'3'=>55,'4'=>56,'5'=>57,'6'=>58,'7'=>59,'8'=>60,
			'9'=>61,'!'=>62,'"'=>63,'%'=>64,'&'=>65,'\''=>66,'*'=>67,'+'=>68,'-'=>69,'.'=>70,
			'/'=>71,'_'=>72,','=>73,':'=>74,';'=>75,'='=>76,'<'=>77,'>'=>78,'?'=>79
		);
		for($i=0; $i<4; $i++) {
			$a=mb_substr($mrc, $i, 1);
			if($s[$a]) {
				$n[$i]=$s[$a];
			}
			else {
				$n[$i]=0;
			}
		}
		$result=$n[1]*80;
		$result=($result+$n[2])*80;
		$result+=$n[0]+$n[3];
		return $result;
	}
}
